"""
Utilities related to "multi-component SMILES", i.e. strings containing multiple compounds
in SMILES notation, which may include fragment bonds.
"""
from functools import partial
from typing import Callable, Iterable, List, Optional

from rxn.utilities.containers import remove_duplicates

from .conversion import canonicalize_smiles


def multicomponent_smiles_to_list(
    multicomponent_smiles: str, fragment_bond: Optional[str] = None
) -> List[str]:
    """
    Convert a string of molecules into a list of molecules (taking fragment bonds into account).

    Args:
        multicomponent_smiles: multicomponent SMILES string to convert to a list.
        fragment_bond: fragment bond.

    Returns:
        The list of molecule SMILES comprised in the multi-component SMILES string.
    """
    molecules = multicomponent_smiles.split(".")
    molecules = [molecule for molecule in molecules if molecule != ""]

    # replace fragment bonds if necessary
    if fragment_bond is not None:
        molecules = [molecule.replace(fragment_bond, ".") for molecule in molecules]
    return molecules


def list_to_multicomponent_smiles(
    molecules: Iterable[str], fragment_bond: Optional[str] = None
) -> str:
    """
    Convert a list of molecules into a string representation (taking fragment
    bonds into account).

    Args:
        molecules: molecule SMILES strings to merge into a multi-component SMILES string.
        fragment_bond: fragment bond.

    Returns:
        A multi-component SMILES string.
    """
    # replace fragment bonds if necessary
    if fragment_bond is not None:
        molecules = [molecule.replace(".", fragment_bond) for molecule in molecules]

    return ".".join(molecules)


def apply_to_multicomponent_smiles(
    multicomponent_smiles: str,
    fn: Callable[[str], str],
    fragment_bond: Optional[str] = None,
) -> str:
    """
    Apply a function to the individual compounds in a multi-component SMILES string.

    Args:
        multicomponent_smiles: multi-component SMILES string to apply the function to.
        fn: function to apply on the distinct molecule SMILES.
        fragment_bond: fragment bond to use when parsing.

    Returns:
        New multi-component SMILES string after application of the function to the molecules.
    """
    molecules = multicomponent_smiles_to_list(
        multicomponent_smiles, fragment_bond=fragment_bond
    )
    molecules = [fn(molecule) for molecule in molecules]
    return list_to_multicomponent_smiles(molecules, fragment_bond=fragment_bond)


def canonicalize_multicomponent_smiles(
    multicomponent_smiles: str,
    fragment_bond: Optional[str] = None,
    check_valence: bool = True,
) -> str:
    """
    Canonicalize the molecules of a multi-component SMILES string.
    """
    canonicalize_fn = partial(canonicalize_smiles, check_valence=check_valence)
    return apply_to_multicomponent_smiles(
        multicomponent_smiles, canonicalize_fn, fragment_bond=fragment_bond
    )


def sort_multicomponent_smiles(multicomponent_smiles: str) -> str:
    """
    Sort the molecule SMILES in a multi-component SMILES string alphabetically.

    Note: no fragment bond is needed here, as it would not have any effect at all.
    """
    return list_to_multicomponent_smiles(
        sorted(multicomponent_smiles_to_list(multicomponent_smiles))
    )


def remove_duplicates_in_multicomponent_smiles(multicomponent_smiles: str) -> str:
    """
    Remove duplicate molecule SMILES strings in a multi-component SMILES string.

    Note: no fragment bond is needed here, as it would not have any effect at all.
    """
    return list_to_multicomponent_smiles(
        remove_duplicates(multicomponent_smiles_to_list(multicomponent_smiles))
    )



import re
from typing import List

SMILES_TOKENIZER_PATTERN = r"(\%\([0-9]{3}\)|\[[^\]]+]|Br?|Cl?|N|O|S|P|F|I|b|c|n|o|s|p|\||\(|\)|\.|=|#|-|\+|\\|\/|:|~|@|\?|>>?|\*|\$|\%[0-9]{2}|[0-9])"
SMILES_REGEX = re.compile(SMILES_TOKENIZER_PATTERN)


class TokenizationError(ValueError):
    """Exception raised in RDKit."""

    def __init__(self, title: str, detail: str):
        """
        Initialize TokenizationError.

        Args:
            title (str): title of the error.
            detail (str): decscription of the error.
        """
        self.type = "TokenizationError"
        self.title = title
        self.detail = detail


def to_tokens(smiles: str) -> List[str]:
    """
    Tokenize a SMILES molecule or reaction into a list of tokens.

    Args:
        smiles: SMILES string to tokenize.

    Raises:
        TokenizationError: in case of mismatch between the SMILES and the joined tokens.

    Returns:
        List of tokens (give back the original SMILES string if appended).
    """
    tokens = [token for token in SMILES_REGEX.findall(smiles)]

    if smiles != "".join(tokens):
        raise TokenizationError(
            "SmilesJoinedTokensMismatch",
            f'SMILES="{smiles}" != joined_tokens="{"".join(tokens)}"',
        )

    return tokens


def tokenize_smiles(smiles: str) -> str:
    """
    Tokenize a SMILES molecule or reaction, and join the tokens with spaces.

    Args:
        smiles: SMILES string to tokenize, for instance 'CC(CO)=N>>CC(C=O)N'.

    Returns:
        SMILES string after tokenization, for instance 'C C ( C O ) = N >> C C ( C = O ) N'.
    """
    tokens = to_tokens(smiles)
    return " ".join(tokens)


def detokenize_smiles(tokenized_smiles: str) -> str:
    """
    Detokenize a tokenized SMILES string (that contains spaces between the characters).

    Args:
        tokenized_smiles: tokenized SMILES, for instance 'C C ( C O ) = N >> C C ( C = O ) N'

    Returns:
        SMILES after detokenization, for instance 'CC(CO)=N>>CC(C=O)N'
    """
    return tokenized_smiles.replace(" ", "")



from typing import Iterable

from rdkit.Chem import CombineMols
from rdkit.Chem.rdchem import Mol


def combine_mols(mols: Iterable[Mol]) -> Mol:
    """Combine multiple RDKit Mols into one.

    RDKit has a function for that but it only supports two mols.

    Args:
        mols: RDKit Mols to combine.
    """

    resulting_mol = Mol()
    for mol in mols:
        resulting_mol = CombineMols(resulting_mol, mol)
    return resulting_mol
